/*
 * Curve Plotter is a Java application/applet allowing to draw
 * an arbitrary 2-dimension parametric curve in
 * the Cartesian coordinate system.
 * 
 * A curve is given as 
 * 
 * x(arg) = <expression>
 * y(arg) = <expression>
 * Min = <expression>
 * Max = <expression>
 * Delta = <expression>
 * X0 - the x-coordinate of the center of the drawing area
 * Y0 - the y-coordinate of the center of the drawing area
 * Scale - the number of screen dots in the coordinate unit
 * 
 * where 
 * 
 * arg is in {t| t in [Min, Max] : Exists k in Z : t = Min+k*Delta}
 * 
 * <expression> ::= {-} <summand> {+ <summand>} | {-} <summand> {- <summand>}
 * <summand>    ::= <multiplicand> {* <multiplicand>} | <multiplicand> {/ <multiplicand>}
 * <term>       ::= <atom> {^ <atom>}
 * <atom>       ::= <number> | arg | <function> ( <expression> ) | ( <expression> )
 * <function>   ::= E | PI | abs | acos | asin | atan | cos | ceil | cos |
 * 						  exp |  log |  sin | sqrt | tan | floor 
 * <number>     ::= <integer> [. <integer> ]	
 * <integer>    ::= <digit> { <digit> }
 * 
 * The application window contains the drawing area,
 * the text field to type x(arg), y(arg), Min, Max, Delta, and
 * buttons to save the current curve, read a saved curve, draw the current curve,
 * clear the drawing area.
 * 
 * A curve is stored in a plain text file containing textual presentation for
 * x(arg), y(arg), Min, Max, Delta, X0, Y0, Scale.
 *   
 * Author.		M.Lebedev (lebedev@cecmow.enet.dec.com)
 * Date.		06.04.97
 *
 */

import java.awt.*;
import java.io.*;
//import java.math.*;
import java.util.*;
import java.applet.*;

public class Graph extends Applet implements Runnable {
	Button startButton = null;
	GraphFrame f = null;
	Thread t = null;

    public static void main(String args[]) {
		Graph g = new Graph();
		g.f = new GraphFrame();
    }

    public void init () {
		startButton = new Button("Start");
        setLayout(new BorderLayout());
		add("Center", startButton);
    }

	public void start() {
		if(t==null) t = new Thread(this);
		t.start();
	}

	public synchronized void run() {
		try {
			wait();
			System.out.println("after wait");
			f = new GraphFrame();
		}
		catch(InterruptedException e) {
		}
	}

    public synchronized boolean action(Event event, Object what) {
        if (event.target instanceof Button) {
            notify();
        } 
        return true;
    }
}

class GraphFrame extends Frame { // implements Runnable {
	Parser parser = null;

	public GraphFrame() {
        super("Graph window");

        //{{INIT_CONTROLS
        setLayout(new BorderLayout());
        addNotify();
        resize(insets().left + insets().right + 511, insets().top + insets().bottom + 604);
        drawPanel=new DrawPanel();
        add("Center", drawPanel);
        drawPanel.reshape(insets().left + 2,insets().top + 2,502,317);
        ControlPanel=new Panel();
        ControlPanel.setLayout(null);
        add("South", ControlPanel);
        ControlPanel.reshape(insets().left + 0,insets().top + 319,504,253);
        xtLabel=new Label("x(arg)");
        ControlPanel.add(xtLabel);
        xtLabel.reshape(7,13,34,25);
        xtEdit=new TextField(53);
        ControlPanel.add(xtEdit);
        xtEdit.reshape(56,13,441,25);
        label1=new Label("y(arg)");
        ControlPanel.add(label1);
        label1.reshape(7,45,34,26);
        ytEdit=new TextField(53);
        ControlPanel.add(ytEdit);
        ytEdit.reshape(56,45,441,26);
        tminLabel=new Label("Min");
        ControlPanel.add(tminLabel);
        tminLabel.reshape(7,78,35,25);
        tminEdit=new TextField(11);
        ControlPanel.add(tminEdit);
        tminEdit.reshape(56,78,98,25);
        tmaxLabel=new Label("Max");
        ControlPanel.add(tmaxLabel);
        tmaxLabel.reshape(7,110,35,26);
        tmaxEdit=new TextField(11);
        ControlPanel.add(tmaxEdit);
        tmaxEdit.reshape(56,110,98,26);
        label2=new Label("Delta");
        ControlPanel.add(label2);
        label2.reshape(7,142,35,27);
        dtEdit=new TextField(11);
        ControlPanel.add(dtEdit);
        dtEdit.reshape(56,142,98,27);
        x0Label=new Label("X0");
        ControlPanel.add(x0Label);
        x0Label.reshape(189,78,28,25);
        x0Edit=new TextField(13);
        ControlPanel.add(x0Edit);
        x0Edit.reshape(231,78,112,25);
        y0Label=new Label("Y0");
        ControlPanel.add(y0Label);
        y0Label.reshape(189,110,28,26);
        y0Edit=new TextField(13);
        ControlPanel.add(y0Edit);
        y0Edit.reshape(231,110,112,26);
        scaleLabel=new Label("Scale");
        ControlPanel.add(scaleLabel);
        scaleLabel.reshape(175,142,42,27);
        scaleEdit=new TextField(13);
        ControlPanel.add(scaleEdit);
        scaleEdit.reshape(231,142,112,27);
        openButton=new Button("Open");
        ControlPanel.add(openButton);
        openButton.reshape(370,78,45,25);
        saveButton=new Button("Save");
        ControlPanel.add(saveButton);
        saveButton.reshape(370,110,45,25);
        drawButton=new Button("Draw");
        ControlPanel.add(drawButton);
        drawButton.reshape(425,78,45,25);
        clearButton=new Button("Clear");
        ControlPanel.add(clearButton);
        clearButton.reshape(425,110,45,25);
        logEdit=new TextArea();
        ControlPanel.add(logEdit);
        logEdit.reshape(5,182,492,68);
        //}}

        //{{INIT_MENUS
        //}}

		xtEdit.setText("5*cos(arg)");
		ytEdit.setText("5*sin(3*arg)");
		x0Edit.setText("0.");
		y0Edit.setText("0.");
		scaleEdit.setText("10");
		tminEdit.setText("0.");
		tmaxEdit.setText("6.");
		dtEdit.setText("0.1");

		show();
    }

    public boolean handleEvent(Event event) {
//		System.out.println(event.toString());
    	if (event.id == Event.ACTION_EVENT && event.target == drawButton) {
    	    	clickedDrawButton();
    	    	return true;
    	}
    	else 
    	if (event.id == Event.ACTION_EVENT && event.target == clearButton) {
    	    	clickedClearButton();
    	    	return true;
    	}
    	else 
    	if (event.id == Event.ACTION_EVENT && event.target == saveButton) {
    	    	clickedSaveButton();
    	    	return true;
    	}
    	else 
    	if (event.id == Event.ACTION_EVENT && event.target == openButton) {
    	    	clickedOpenButton();
    	    	return true;
    	}
    	else 

    	if (event.id == Event.WINDOW_DESTROY) {
    	    System.exit(0);
    	    return true;
    	}
    	return super.handleEvent(event);
    }

    //{{DECLARE_MENUS
    //}}

    //{{DECLARE_CONTROLS
    DrawPanel drawPanel;
    Panel ControlPanel;
    Label xtLabel;
    TextField xtEdit;
    Label label1;
    TextField ytEdit;
    Label tminLabel;
    TextField tminEdit;
    Label tmaxLabel;
    TextField tmaxEdit;
    Label label2;
    TextField dtEdit;
    Label x0Label;
    TextField x0Edit;
    Label y0Label;
    TextField y0Edit;
    Label scaleLabel;
    TextField scaleEdit;
    Button openButton;
    Button saveButton;
    Button drawButton;
    Button clearButton;
    TextArea logEdit;
    //}}
    public void clickedOpenButton() {
        FileDialog fd
		= new FileDialog(
			this,
			"Open file with curve",
			FileDialog.LOAD);

	    fd.show();
	    if( fd.getFile() != null ) {
    		DataInputStream in;
    		FileInputStream fis = null;
			try {
    			fis = new FileInputStream(fd.getDirectory()+fd.getFile());
    			in = new DataInputStream(new BufferedInputStream(fis));
				xtEdit.setText(in.readLine());
				ytEdit.setText(in.readLine());
				x0Edit.setText(in.readLine());
				y0Edit.setText(in.readLine());
				scaleEdit.setText(in.readLine());
				tminEdit.setText(in.readLine());
				tmaxEdit.setText(in.readLine());
				dtEdit.setText(in.readLine());

                fis.close();
                in.close();
    		}
    		catch(FileNotFoundException e) {
    			log("FileNotFoundException: " + e.getMessage());
    		}
    		catch(IOException e) {
    			log("IOException: " + e.getMessage());
   			}
		}
    }
    public void clickedSaveButton() {
        FileDialog fd
		= new FileDialog(
			this,
			"Save curve nto a file",
			FileDialog.SAVE);

		fd.show();
		if( fd.getFile() != null ) {
   			FileOutputStream fos = null;
   			PrintStream out = null;
			try {
    			fos = new FileOutputStream(fd.getDirectory()+fd.getFile());
 				out = new PrintStream(new BufferedOutputStream(fos));

				out.println(xtEdit.getText());
				out.println(ytEdit.getText());
				out.println(x0Edit.getText());
				out.println(y0Edit.getText());
				out.println(scaleEdit.getText());
				out.println(tminEdit.getText());
				out.println(tmaxEdit.getText());
				out.println(dtEdit.getText());

                out.close();
				fos.close();
			}
			catch(FileNotFoundException e) {
    			log("FileNotFoundException: " + e.getMessage());
    		}
    		catch(IOException e) {
    			log("IOException: " + e.getMessage());
   			}
		}
    }
    public void clickedDrawButton() {
		if(drawButton.getLabel().equals("Stop")) {
			stopDraw();
		} else {
			drawButton.setLabel("Stop");
			try {
				drawPanel.onDraw(
					scaleEdit.getText(),
					x0Edit.getText(),
					y0Edit.getText(),
					tminEdit.getText(),
					tmaxEdit.getText(),
					dtEdit.getText(),
					xtEdit.getText(),
					ytEdit.getText()
				);
			}
			catch(MyException e) {
				log("Parsing. MyException: " + e.getMessage());
			}
			catch(NumberFormatException e) {
				log("Parsing. NumberFormatException: " + e.getMessage());
			}
			drawButton.setLabel("Draw");
		}
	}
    public void clickedClearButton() {
		drawPanel.icurve = null;
		drawPanel.repaint();
	}
	private void stopDraw() {
		drawButton.setLabel("Draw");
		drawPanel.icurve = null;
	}
	protected void log(String str)
    {
        logEdit.appendText(str + "\n");
    }
}

class DrawPanel extends Canvas {
	public double tmin, tmax, dt;
	public R rmid = null;
	R irmid = null;
	public int scale;
	Dimension d = null;
	int color = 9;

	static final Color colors[] = {
		Color.black,
		Color.blue, 
		Color.cyan, 
		Color.darkGray,
		Color.gray, 
		Color.green, 
		Color.lightGray, 
		Color.magenta, 
		Color.orange, 
		Color.pink, 
		Color.red, 
		Color.yellow
	};

	public Curve icurve = null;

	R ri0 = null;
	
	public DrawPanel() {
		setBackground(Color.white);
    }

	R r2ir(R r) {
		return r.sub(ri0).mul(scale);
	}

	R ri2r(R ir) {
		return ir.mul((double)(1./scale)).add(ri0);
	}
    
	public void onDraw(
		String scaleText,
		String x0Text,
		String y0Text,
		String tminText,
		String tmaxText,
		String dtText,
		String xtText,
		String ytText) throws MyException
	{
		scale = Integer.valueOf(scaleText).intValue();
		rmid = new R(Double.valueOf(x0Text).doubleValue(),
					 Double.valueOf(y0Text).doubleValue());
		
		Parser parser = new Parser();
		Calculator c = new Calculator();

		tmin = c.calculate(parser.parse(tminText), 0.);
		tmax = c.calculate(parser.parse(tmaxText), 0.);
		dt   = Double.valueOf(dtText).doubleValue();

		String programmX = parser.parse(xtText);
		String programmY = parser.parse(ytText);
		icurve = new Curve(programmX, programmY);

		repaint();
    }

	public void paint(Graphics g) {
		if( icurve != null ) {
			drawGraph(g);
		} else {
			d = size();
			g.setColor(getBackground());
			g.setPaintMode();
			g.clearRect(0, 0, d.width, d.height);
		}
	}

	public void update(Graphics g) {
		if( icurve != null ) {
			drawGraph(g);
		} else {
			d = size();
			g.setColor(getBackground());
			g.setPaintMode();
			g.clearRect(0, 0, d.width, d.height);
		}
	}
/*
    public boolean handleEvent(Event event) {
		System.out.println(event.toString());
		return true;
	}

	public synchronized void run() {
		repaint();
		try {
			wait();
		}
		catch(InterruptedException e) {
		}
	}
*/	
    public void drawGraph(Graphics g) {
		d = size();
		irmid = new R(d.width/2, -d.height/2);
		ri0 = rmid.sub(irmid.mul((double)(1./scale)));		

		color = (color + 1) % colors.length;
		if(colors[color].equals(getBackground()) || colors[color].equals(Color.blue))  {
			color = (color + 1) % colors.length;
		}
		
		g.setColor(Color.blue);
		g.setPaintMode();

//	draw axes...

		R ir0 = r2ir (new R(0., 0.)); 

		int ix0 = (int)ir0.x;
		int iy0 = -(int)ir0.y; 

		g.drawLine(ix0, 0, ix0, d.height);
		g.drawLine(0, iy0, d.width, iy0);

		int ix, iy;
		for(int i = 1; true; i++) {
			int ixmin = ix0-i*scale;
			int ixmax = ix0+i*scale;
			int iymin = iy0-i*scale;
			int iymax = iy0+i*scale;
			g.drawLine(ix0-2, iymax, ix0+2, iymax);
			g.drawLine(ix0-2, iymin, ix0+2, iymin);
			g.drawLine(ixmax, iy0-2, ixmax, iy0+2);
			g.drawLine(ixmin, iy0-2, ixmin, iy0+2);

			boolean b =	(ixmin < 0 || ixmin > d.width) &&
						(ixmax < 0 || ixmax > d.width) &&
						(iymax < 0 || iymax > d.height) &&
						(iymin < 0 || iymax > d.height);

			if(b) break;
		}

		g.setColor(colors[color]);

		drawCurve(g, tmin);
    }

	protected void drawCurve(Graphics g, double tt) {
		if(tt > tmax) return;

		double t = tt;

		try {
			R r = new R(icurve.x(t), icurve.y(t));
			R irnew = r2ir(r);
			R ir = irnew;
			for(t=t+dt; t <= tmax; t+=dt) {
				if(!isInside(irnew)) throw new ArithmeticException("Outside view");
				irnew = r2ir(r.setR(icurve.x(t), icurve.y(t)));
				g.drawLine((int)ir.x, -(int)ir.y, (int)irnew.x, -(int)irnew.y);
				ir = irnew;
			}
		}
		catch(MyException e) {
			System.out.println("Exception: "+e.getMessage());
			return;
		}
		catch(ArithmeticException  e) {
			drawCurve(g, t+dt);
		}
	}

	protected boolean isInside(R ir) {
		return (0<=ir.x && ir.x<=d.width) && (0<=-ir.y && -ir.y<=d.height);
	}
}

class R {
	public double x;
	public double y;

	public R() {
		x=0.;
		y=0.;
	}

	public R(double anX, double anY) {
		x=anX;
		y=anY;
	}

	public R(int anX, int anY) {
		x=(double)anX;
		y=(double)anY;
	}

	public R(R r) {
		x=r.x;
		y=r.y;
	}

	public R setR(double anX, double anY) {
		x=anX;
		y=anY;
		return this;
	}

	public R add(R r) {
		return new R(x+r.x, y+r.y);
	}

	public R sub(R r) {
		return new R(x-r.x, y-r.y);
	}

	public R mul(double s) {
		return new R(s*x, s*y);
	}
}

class Lex {
	public static final int NULL		= -2;
	public static final int EOL			= -1;

	public static final int PLUS		= 0;
	public static final int MINUS		= 1;
	public static final int MUL			= 2;
	public static final int DIV			= 3;
	public static final int POW			= 4;
	public static final int LEFT		= 5;
	public static final int RIGHT		= 6;
	public static final int ARG			= 7;
	public static final int e			= 8;
	public static final int pi			= 9;

	public static final int ABS			= 10;
	public static final int ACOS		= 11;
	public static final int ASIN		= 12;
	public static final int ATAN		= 13;
	public static final int CEIL		= 14;
	public static final int COS			= 15;
	public static final int EXP			= 16;
	public static final int FLOOR		= 17;
	public static final int LOG			= 18;
	public static final int SIN			= 19;
	public static final int SQRT		= 20;
	public static final int TAN			= 21;

	public static final int UMINUS		= 22;

	public static final int NUM			= 23;

	public String lexem = null;
	public double number;
	
	static final String [] lexems = {
		"+",
		"-",
		"*",
		"/",
		"^",
		"(",
		")",
		"arg",
		"E", 
		"PI", 
		"abs",
		"acos",
		"asin",
		"atan",
		"ceil",
		"cos",
		"exp",
		"floor",
		"log",
		"sin",
		"sqrt",
		"tan",	
		"u-"	
	};

	String string = null;
	
	public Lex(String str) {
		string = str;
	}

	int getLex() throws MyException {
		int lexcode = -1;
		int lexlen = 0;
		string = string.trim();

		if(string.equals("")) return -1;

		for(int i=0; i<lexems.length; i++) {
			if(string.startsWith(lexems[i])) {
				lexcode=i;
				lexlen=lexems[i].length();
				break;
			}
		}
		if(lexcode==-1) {
			if(Character.isDigit(string.charAt(0))) {
				int i;
				for(i=0; i<string.length() &&
					(Character.isDigit(string.charAt(i)) ||
					string.charAt(i)=='.'); i++);
				lexlen = i;
				lexcode = NUM;
				number = (Double.valueOf(string.substring(0, lexlen))).doubleValue();
			} else {
				throw new MyException("Unknown lexem, " + string);
			}
		}
		lexem = string.substring(0, lexlen);
		string = string.substring(lexlen, string.length());
		return lexcode;
	}
}

class Parser {
	Lex lx = null;
	String programm = null;
	public Parser() {}
/*
 *
 *	Grammar:
 *
 *		S : {-} T {+ T} | {-} T {- T}
 *		T : P {* P} | P {/ P}
 *		P : E {^ E}
 *		E : number | arg | func ( S ) | ( S )
 *		func : sin | cos | ...
 *		number : double
 *
 *	post-parse(str) == (str is valid => programm == postfix form of str)
 */
	public String parse(String str) throws MyException {
		lx = new Lex(str);
		programm = "";
		if(parseS() != Lex.EOL) throw new MyException("EOL expected.");
//		System.out.println(programm);
		return programm;
	}

	protected int parseS() throws MyException {
		int lex = lx.getLex();
		if(lex==Lex.MINUS) {
			lex = parseT(Lex.NULL);
			programm += "u-" + " ";
		} else {
			lex = parseT(lex);
		}
		while(true) {
			if(lex!=Lex.PLUS && lex!=Lex.MINUS) return lex;
			String op = lx.lexem;
			lex = parseT(Lex.NULL);
			programm += op + " ";
		}
	}
	
	protected int parseT(int lexem) throws MyException {
		int lex = parseP(lexem);
		while(true) {
			if(lex!=Lex.MUL && lex!=Lex.DIV) return lex;
			String op = lx.lexem;
			lex = parseP(Lex.NULL);
			programm += op + " ";
		}
	}
	
	protected int parseP(int lexem) throws MyException {
		int lex = parseE(lexem);
		while(true) {
			if(lex!=Lex.POW) return lex;
			String op = lx.lexem;
			lex = parseE(Lex.NULL);
			programm += op + " ";
		}
	}
	
	protected int parseE(int lexem) throws MyException {
		int lex = lexem;
		if( lex==Lex.NULL ) lex = lx.getLex();
		switch(lex) {
			case Lex.ARG:
			case Lex.NUM:
			case Lex.pi:
			case Lex.e:
				programm += lx.lexem + " ";
			break;
			case Lex.LEFT:
				lex = parseS();
				if(lex!=Lex.RIGHT) {
					throw new MyException(") expected");
				}
			break;
			case Lex.ABS:
			case Lex.ACOS:
			case Lex.ASIN:
			case Lex.ATAN:
			case Lex.CEIL:
			case Lex.COS:
			case Lex.EXP:
			case Lex.FLOOR:
			case Lex.LOG:
			case Lex.SIN:
			case Lex.SQRT:
			case Lex.TAN:
				String op = lx.lexem;
				lex=lx.getLex();
				if(lex!=Lex.LEFT) {
					throw new MyException("( expected");
				}
				lex = parseS();
				if(lex!=Lex.RIGHT) {
					throw new MyException(") expected");
				}
				programm += op + " ";
			break;
			case Lex.EOL:
				throw new MyException("Unexpected end");
			default:
				throw new MyException("Unexpected " + lx.lexem);
		}
		lex = lx.getLex();
		return lex;
	}
}

class MyException extends Exception {
	public MyException(String str) {
		super(str);
	}
}

class Curve extends Calculator{
	String programmX = null;
	String programmY = null;
	
	public Curve(String prgX, String prgY) {
		super();
		programmX = prgX;	
		programmY = prgY;
	}

	public double x(double t) throws MyException {
		return calculate(programmX, t);
	}

	public double y(double t) throws MyException {
		return calculate(programmY, t);
	}
}

class Calculator extends Stack {
	public Calculator(){
	}

	public double _pop () {
		return ((Double)pop()).doubleValue();
	}

	public double calculate(String prg, double t) throws MyException {
		Lex lx = new Lex(prg);

		empty();
		
		boolean cont = true;
		while(cont) {
			int lex = lx.getLex();
			switch(lex) {
				case Lex.ARG: 
					push(new Double(t));
				break;
				case Lex.e: 
					push(new Double(Math.E));
				break;
				case Lex.pi: 
					push(new Double(Math.PI));
				break;
				case Lex.ABS: 
					push(new Double(Math.abs(_pop())));
				break;
				case Lex.ACOS: 
					push(new Double(Math.acos(_pop())));
				break;
				case Lex.ASIN: 
					push(new Double(Math.asin(_pop())));
				break;
				case Lex.ATAN: 
					push(new Double(Math.atan(_pop())));
				break;
				case Lex.CEIL: 
					push(new Double(Math.ceil(_pop())));
				break;
				case Lex.COS: 
					push(new Double(Math.cos(_pop())));
				break;
				case Lex.EXP: 
					push(new Double(Math.exp(_pop())));
				break;
				case Lex.FLOOR: 
					push(new Double(Math.floor(_pop())));
				break;
				case Lex.LOG: 
					push(new Double(Math.log(_pop())));
				break;
				case Lex.SIN: 
					push(new Double(Math.sin(_pop())));
				break;
				case Lex.SQRT: 
					push(new Double(Math.sqrt(_pop())));
				break;
				case Lex.TAN: 
					push(new Double(Math.tan(_pop())));
				break;
				case Lex.NUM:
					push(new Double(lx.number));
				break;
				case Lex.PLUS: 
					push(new Double(_pop()+_pop()));
				break;
				case Lex.MINUS: 
					push(new Double(-_pop()+_pop()));
				break;
				case Lex.MUL: 
					push(new Double(_pop()*_pop()));
				break;
				case Lex.DIV: 
					push(new Double((1./_pop())*_pop()));
				break;
				case Lex.POW:
					double pwr = _pop(); 
					push(new Double(Math.pow(_pop(),pwr)));
				break;
				case Lex.UMINUS:
					push(new Double(-_pop()));
				break;
				case Lex.EOL:
					cont = false;
				break;
			}
		}
		return _pop();
	}
}

